#ifndef SOLVER_HPP_
#define SOLVER_HPP_


#include "array_queue.hpp"
#include "array_priority_queue.hpp"
#include "array_set.hpp"
#include "array_map.hpp"


namespace ics {


//Given a solution (a queue of operators that when applied on the start state generate the stop
//  state), return the solution as a nicely formatted std::string.
std::string illustrate_solution (ics::State state, ics::ArrayQueue<ics::Operator> operators) {
  if (operators.empty())
    return "No/Trivial solution found";

  std::ostringstream solution;
  solution << "Solution length = " << operators.size() << "\n  " << state  << ": Starting State\n";
  try {
    for (ics::Operator op : operators) {
      state = op.apply(state);
      solution << "  " << state << ": " << op << "\n";
    }
    solution<< "  " << state << ": Final State";
  } catch (const std::exception& e) {
    solution << e.what();
  }
  return solution.str();
}



//Find a solution to a problem and return it.
//The actual data type that matches the Queue parameter "exploring" might be a queue
//  (see breadth_first_solution) or a priority queue (best_first_solution): the algorithm for
//  the solution is the same, but depends on how dequeue is implemented.
template<class Queue>
ics::ArrayQueue<ics::Operator> solve_it (ics::Problem& problem,
                                         Queue&& exploring) {
  int operator_count = 0;
  ics::State start   = problem.get_start_state();
  ics::State stop    = problem.get_stop_state();

  //Trivial solutions: no operators needed
  if (stop == start) {
    std::cout << "Found Solution: Operator applications = 0 (start state = stop state)" << std::endl;
    return ics::ArrayQueue<ics::Operator>();
  }

  ics::ArraySet<ics::Operator> operators(problem.get_all_operators());

//  ics::ArrayMap<ics::State,ics::ArrayQueue<ics::Operator>> solutions;            //state -> ics::ArrayQueue[Operator]
  ics::ArrayMap<ics::State,ics::ArrayQueue<ics::Operator>>& solutions = problem.solutions;              //state -> ics::ArrayQueue[Operator]
  solutions.clear();

  exploring.enqueue(start);
  solutions[start] = ics::ArrayQueue<ics::Operator>();                             //initial state -> no operations

  //Are there still states to explore
  while(!exploring.empty()) {
    ics::State current_state = exploring.dequeue();                                //state to explore next

    //Try all operators and examine the state they lead to for uniqueness
    for (ics::Operator op : operators) {
      operator_count++;

      try {
        ics::State new_state = op.apply(current_state);

        if (!solutions.has_key(new_state)) {
          //extend solution to include this operator
          ics::ArrayQueue<ics::Operator> current_solution = solutions[current_state];  //operators to get to this state
          ics::ArrayQueue<ics::Operator> new_solution(current_solution);
          new_solution.enqueue(op);

          //if stop state, return solution
          if (new_state == stop) {
            std::cout << "Found Solution: Operator applications = " << operator_count
                      << "/states reach = " << (solutions.size()+1) << std::endl;
            return new_solution;
          }

          //update solutions map and exploring queue
          solutions[new_state] = new_solution;
          exploring.enqueue(new_state);
        }
      } catch (const std::exception& e) {
        //Just apply next operator if this one fails to work; i.e., skip it
      }

    }
  }

  //Failed to find solution; return an empty queue of operators.
  //Note that this is the same result as the trivial solution (above).
  std::cout << "No Solution: Operator applications = " << operator_count
            << "/states reach = " << (solutions.size()+1) << std::endl;
  return ics::ArrayQueue<ics::Operator>();
}






//Use breadth-first searching to find an "optimal" solution (the one
//  with the minimum number of operators needed to transform the initial
//  state to the final state).
ics::ArrayQueue<ics::Operator> breadth_first_solution (ics::Problem& problem) {
  return solve_it(problem, ics::ArrayQueue<ics::State>());
}



//Use the gt function (state i has higher priority than state j if i is heuristically closer to
//  the stop state) to try to find a solution faster (search the tree, looking at fewer internal
//  states), but typically not find an optimal solution.
//The above is a pure best-first algorithm. If the gt function computes its result by adding
//  (a) teh number of operators it takes to get to a state, and
//  (b) the heuristic does not over-estimate how many operators it takes to reach the next state
//  then the result will be an "optimal" solution, typically found by examining more operators
//  than a pure best-first search but fewer than a breadth-first search.
ics::ArrayQueue<ics::Operator> best_first_solution (ics::Problem& problem,
                                                    bool (*gt)(const ics::State& a, const ics::State& b)) {
  return solve_it(problem, ics::ArrayPriorityQueue<ics::State>(gt));
}

}

#endif /* SOLVER_HPP_ */
