#include "water_jugs.hpp"


namespace ics {


std::string to_string(int x) {
  std::ostringstream ss;
  ss << x;
  return ss.str();
}

State::State()
  : jug_state(nullptr), problem(nullptr)  {
}


State::State(const State& s)
  : problem(s.problem) {
  jug_state = new int[problem->length];
  for (int i = 0; i < problem->length; ++i)
    jug_state[i] = s.jug_state[i];

}
State::State(int jugs[], Problem* prob)
  : problem(prob) {
  for (int i = 0; i < problem->length; ++i)
    if (jugs[i] < 0 || jugs[i] > problem->jug_sizes[i])
      throw "State::State: size of jug" + to_string(i) + "is " + to_string(jugs[i]) +
            "must be >= 0 and <= " + to_string(problem->jug_sizes[i]);
  jug_state = new int[problem->length];
  for (int i = 0; i < problem->length; ++i)
    jug_state[i] = jugs[i];
}


State::~State() {
  if (jug_state != nullptr) {
    delete[] jug_state;
  }
}


int State::how_close() const {
//  int answer = 0;                                  //f(x) = h(x)
  int answer = problem->solutions[*this].size();   //f(x) = #ops + h(x) : the a* algorithm
  for (int i=0; i<problem->length; ++i)
     answer += (jug_state[i] == problem->stop.jug_state[i] ? 0 : 1);
     //answer += abs(jug_state[i]-s.jug_state[i]);                              //Alternative
     //answer += (jug_state[i]-s.jug_state[i]) * (jug_state[i]-s.jug_state[i]); //Alternative
  return answer;
}

State&  State::operator = (const State& rhs) {
  if (this == &rhs)
    return *this;
  problem = rhs.problem;
  if (jug_state != nullptr)
    delete [] jug_state;
  jug_state = new int[problem->length];
  for (int i = 0; i < problem->length; ++i)
    jug_state[i] = rhs.jug_state[i];
  return *this;
}


bool State::operator == (const State& rhs) const {
  if (this == &rhs)
    return true;
  if (problem->length != rhs.problem->length)
    return false;
  for (int i = 0; i < problem->length; ++i)
    if (jug_state[i] != rhs.jug_state[i])
      return false;
  return true;
}


std::ostream& operator << (std::ostream& outs, const State& s) {
//  outs << "State:" << &s <<  "(" << s.jug_state << ")" << "[jug_state=" << "[";
  outs << "State[jug_state=" << "[";
  for (int i = 0; i < s.problem->length; ++i)
    outs << (i==0 ? "" : ",") << s.jug_state[i];
  outs << "]]";
  return outs;
}



Operator::Operator()
  : from(-1), to(-1) {
}



Operator::Operator(int jug_from, int jug_to, int jug_count) {
  if (jug_from<0 || jug_from>=jug_count)
    throw "Operator::Operator: Illegal from("+to_string(jug_from)+") jug #";
  if (jug_to  <0 || jug_to  >=jug_count)
    throw "Operator::Operator: Illegal to("+to_string(jug_to)+") jug #";

  from = jug_from;
  to   = jug_to;
}


Operator::~Operator() {
}


State Operator::apply (const State& s) const {
  State result_state(s.jug_state,s.problem);
  int to_capacity = s.problem->jug_sizes[to] - s.jug_state[to];
  if (s.jug_state[from] <= to_capacity) {
    //Empty from (becomes 0) into to (incremented)
    result_state.jug_state[to]   += s.jug_state[from];
    result_state.jug_state[from] =  0;
  } else {
    //fill to (becomes full) from from (decremented)
    result_state.jug_state[from] -= to_capacity;
    result_state.jug_state[to]   =  s.problem->jug_sizes[to];
  }
  return result_state;
}


bool Operator::operator == (const Operator& rhs) const {
  if (this == &rhs)
    return true;
  return from == rhs.from && to == rhs.to;
}


std::ostream& operator << (std::ostream& outs, const Operator& o) {
  outs << "pour from jug " << to_string(o.from+1) << " to " << to_string(o.to+1);
//  outs << to_string(o.from+1) << + "->" << to_string(o.to+1); // Shortened version
  return outs;
}




Problem::Problem(int all_jug_sizes[], int start_jugs[], int stop_jugs[], int num_jugs) {
  length = num_jugs;
  jug_sizes = new int[num_jugs];
  for (int i = 0; i < num_jugs; ++i)
    jug_sizes[i] = all_jug_sizes[i];
  start = State(start_jugs,this);
  stop  = State(stop_jugs, this);
}


Problem::~Problem() {
  delete[] jug_sizes;
}


State Problem::get_start_state () const {
  return start;
}

State Problem::get_stop_state () const {
  return stop;
}


ics::ArraySet<Operator> Problem::get_all_operators () const {
  ics::ArraySet<Operator> operators;
  //Transfer i->j and j->i  for all i != j
  for (int i = 0; i < length; ++i)
    for (int j = i+1; j < length; ++j) {
      operators.insert(Operator(i, j, length));
      operators.insert(Operator(j, i, length));
    }

  return operators;
}


std::ostream& operator << (std::ostream& outs, const Problem& p) {
  outs <<  "Problem[jug_sizes=[";
  for (int i=0; i<p.length; ++i)
    outs << (i==0 ? "" : ",") << p.jug_sizes[i];
  outs << "],start=" << p.start << ",stop=" << p.stop << "]";
  return outs;
}




}


